// OpenGL ES 3.1 Fractals fragment shader.
// Written by Nils Liaaen Corneliusen.
// Please refer to the article at http://www.ignorantus.com/pages/gpu_hacks/ for more information.
// Based on code by John Tsiombikas: http://nuclear.mutantstargoat.com/articles/sdr_fract/
#version 310 es
precision highp float;
layout(location = 0) in vec2 v_texCoord;
layout(location = 0) out vec4 outColor;

#define MAXITER 256
#define DEPTH 32

layout(binding = 0) uniform stuff
{
    float sx, sy, sz;
    float re, im;
    float mandel;
};

// Hue table with black at end instead of wrapping
const vec3 cols[7] = vec3[7](
    vec3( 1.0f, 0.0f, 0.0f ),
    vec3( 1.0f, 1.0f, 0.0f ),
    vec3( 0.0f, 1.0f, 0.0f ),
    vec3( 0.0f, 1.0f, 1.0f ),
    vec3( 0.0f, 0.0f, 1.0f ),
    vec3( 1.0f, 0.0f, 1.0f ),
    vec3( 0.0f, 0.0f, 0.0f )
);

vec3 getRGB( float h )
{
    float Slice     = 6.0f * h;
    float SliceInt  = floor( Slice );
    float SliceFrac = Slice - SliceInt;

    int i = int( SliceInt );
    return mix( cols[i], cols[i+1], SliceFrac );
}

void main()
{
    vec4  xv0 = vec4( 0.0f, 0.0f, 0.0f, 1.7778f * (v_texCoord.x - 0.5f) / (0.5f * sz) + sx );
    vec4  yv0 = vec4( 0.0f, 0.0f, 0.0f,           (v_texCoord.y - 0.5f) / (0.5f * sz) + sy );
    bvec4 bv[DEPTH/4];

    float fre = mandel != 0.0f ? xv0.w : re;
    float fim = mandel != 0.0f ? yv0.w : im;

    int i;
    for( i = 0; i < MAXITER; i += DEPTH ) {

        for( int j = 0; j < DEPTH/4; j++ ) {

            xv0.x = xv0.w * xv0.w - yv0.w * yv0.w + fre; yv0.x = 2.0f * xv0.w * yv0.w + fim;
            xv0.y = xv0.x * xv0.x - yv0.x * yv0.x + fre; yv0.y = 2.0f * xv0.x * yv0.x + fim;
            xv0.z = xv0.y * xv0.y - yv0.y * yv0.y + fre; yv0.z = 2.0f * xv0.y * yv0.y + fim;
            xv0.w = xv0.z * xv0.z - yv0.z * yv0.z + fre; yv0.w = 2.0f * xv0.z * yv0.z + fim;

            bv[j].x = xv0.x * xv0.x + yv0.x * yv0.x > 4.0f;
            bv[j].y = xv0.y * xv0.y + yv0.y * yv0.y > 4.0f;
            bv[j].z = xv0.z * xv0.z + yv0.z * yv0.z > 4.0f;
            bv[j].w = xv0.w * xv0.w + yv0.w * yv0.w > 4.0f;

        }

        if( any( bv[0] ) || any( bv[1] ) || any( bv[2] ) || any( bv[3] ) ||
            any( bv[4] ) || any( bv[5] ) || any( bv[6] ) || any( bv[7] ) ) break;

    }

    uint bits = 0u;
    for( int j = 0; j < DEPTH/4; j++ )
        bits |= (uint(bv[j].x)<<(j*4)) | (uint(bv[j].y)<<(j*4+1)) | (uint(bv[j].z)<<(j*4+2)) | (uint(bv[j].w)<<(j*4+3));

    int lsb = findLSB( bits );
    i += lsb == -1 ? 0 : lsb;

    outColor = vec4( getRGB( float(i) / float(MAXITER) ), 1.0f );
}
